/* 
 * leases.c -- tools to manage DHCP leases 
 * Russ Dill <Russ.Dill@asu.edu> July 2001
 */

#include <time.h>
#include <string.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>

#include "debug.h"
#include "dhcpd.h"
#include "files.h"
#include "options.h"
#include "leases.h"
#include "arpping.h"

unsigned char blank_chaddr[] = {[0 ... 15] = 0};

/* clear every lease out that chaddr OR yiaddr matches and is nonzero */
int clear_lease(u_int8_t *chaddr, u_int32_t yiaddr, int ifid)
{
	unsigned int i, j;
	int found_reserved = 0;
	for (j = 0; j < 16 && !chaddr[j]; j++);
	
	for (i = 0; i < server_config[ifid].max_leases; i++)
	{
		if ((j != 16 && !memcmp(leases[i].chaddr, chaddr, 16)) ||
		    (yiaddr && leases[i].yiaddr == yiaddr))
		{
			if(leases[i].expires > server_config[ifid].lease+time(0))
				found_reserved = 1;
			memset(&(leases[i]), 0, sizeof(struct dhcpOfferedAddr));
		}
	}
	return found_reserved;
}

#ifdef RONSCODE
/* add a lease into the table, clearing out any old ones */
struct dhcpOfferedAddr *add_lease(u_int8_t *chaddr, u_int32_t yiaddr, unsigned long lease, int ifid,char *hostname,char *cid)
#else
struct dhcpOfferedAddr *add_lease(u_int8_t *chaddr, u_int32_t yiaddr, unsigned long lease, int ifid)
#endif
{
	struct dhcpOfferedAddr *oldest;
	int found_reserved;
	
	DEBUG(LOG_INFO, "enter: "IPSTR","MACSTR", lease = %lu",
				IP2STR(yiaddr), MAC2STR(chaddr), lease);
	DEBUG(LOG_INFO, "cid = %02X %02X %02X %02X %02X %02X %02X", 
		(unsigned char)cid[0], (unsigned char)cid[1], (unsigned char)cid[2], 
		(unsigned char)cid[3], (unsigned char)cid[4], (unsigned char)cid[5], 
		(unsigned char)cid[6]);

	/* clean out any old ones */
	found_reserved = clear_lease(chaddr, yiaddr, ifid);

	DEBUG(LOG_INFO, "this is %sa reserved node", found_reserved?"":"not ");

	oldest = oldest_expired_lease(ifid);
	
	if (oldest) {
		memcpy(oldest->chaddr, chaddr, 16);
		oldest->yiaddr = yiaddr;
		if(found_reserved)
			oldest->expires = time(0) + 31536000;
		else
			oldest->expires = time(0) + lease;
#ifdef RONSCODE
		memcpy(oldest->hostname,hostname,MAXHOSTNAMELEN);
		memcpy(oldest->cid,cid,16);
		
#endif
	}
	
	return oldest;
}


/* true if a lease has expired */
int lease_expired(struct dhcpOfferedAddr *lease)
{
	return (lease->expires < (unsigned long) time(0));
}	


/* Find the oldest expired lease, NULL if there are no expired leases */
struct dhcpOfferedAddr *oldest_expired_lease(int ifid)
{
	struct dhcpOfferedAddr *oldest = NULL;
	unsigned long oldest_lease = time(0);
	unsigned int i;

	
	for (i = 0; i < server_config[ifid].max_leases; i++) {
		if (oldest_lease > leases[i].expires) {
			oldest_lease = leases[i].expires;
			oldest = &(leases[i]);
		}
	}
#ifdef DHCPD_DEBUG
	if (oldest)
		DEBUG(LOG_INFO, "use oldest lease: "IPSTR","MACSTR"",
				IP2STR(oldest->yiaddr), MAC2STR(oldest->chaddr));
#endif
	return oldest;
		
}


/* Find the first lease that matches chaddr, NULL if no match */
struct dhcpOfferedAddr *find_lease_by_chaddr(u_int8_t *chaddr, int ifid)
{
	unsigned int i;

	DEBUG(LOG_INFO, "chaddr = "MACSTR"", MAC2STR(chaddr));
	for (i = 0; i < server_config[ifid].max_leases; i++)
	{
		if (!memcmp(leases[i].chaddr, chaddr, 16)) {
			DEBUG(LOG_INFO, "lease found: "IPSTR","MACSTR"",
				IP2STR(leases[i].yiaddr), MAC2STR(leases[i].chaddr));
			return &(leases[i]);
		}
	}
	return NULL;
}

struct dhcpOfferedAddr *find_lease_by_cid(u_int8_t *cid, int ifid)
{
	unsigned int j;

	DEBUG(LOG_INFO, "cid = %02X %02X %02X %02X %02X %02X %02X", 
		(unsigned char)cid[0], (unsigned char)cid[1], (unsigned char)cid[2], 
		(unsigned char)cid[3], (unsigned char)cid[4], (unsigned char)cid[5], 
		(unsigned char)cid[6]);

	for (j = 0; j < server_config[ifid].max_leases; j++)
	{
		if (!memcmp(leases[j].cid, cid, 7)) {
			DEBUG(LOG_INFO, "lease found: "IPSTR","MACSTR"",
				IP2STR(leases[j].yiaddr), MAC2STR(leases[j].chaddr));
			return &(leases[j]);
		}
	}
	
	return NULL;
}


/* Find the first lease that matches yiaddr, NULL is no match */
struct dhcpOfferedAddr *find_lease_by_yiaddr(u_int32_t yiaddr, int ifid)
{
	unsigned int i;
	
	DEBUG(LOG_INFO, "yiaddr = "IPSTR"", IP2STR(yiaddr));
	for (i = 0; i < server_config[ifid].max_leases; i++)
		if (leases[i].yiaddr == yiaddr) {
			DEBUG(LOG_INFO, "lease found: "IPSTR","MACSTR"",
				IP2STR(leases[i].yiaddr), MAC2STR(leases[i].chaddr));
			return &(leases[i]);
		}
	
	return NULL;
}


/* find an assignable address, it check_expired is true, we check all the expired leases as well.
 * Maybe this should try expired leases by age... */
u_int32_t find_address(int check_expired, int ifid) 
{
	u_int32_t addr, ret;
	struct dhcpOfferedAddr *lease = NULL;		
	addr = ntohl(server_config[ifid].start); /* addr is in host order here */
	for (;addr <= ntohl(server_config[ifid].end); addr++) {
		/* ie, 192.168.55.0 */
		if (!(addr & 0xFF)) continue;

		/* ie, 192.168.55.255 */
		if ((addr & 0xFF) == 0xFF) continue;
		/* Ron */
		/* ie, this ip is server ip */
		if (addr == ntohl(server_config[ifid].server)) continue;
		/* Ron */
		/* lease is not taken */
		ret = htonl(addr);
		if ((!(lease = find_lease_by_yiaddr(ret, ifid)) ||

		     /* or it expired and we are checking for expired leases */
		     (check_expired  && lease_expired(lease))) &&

		     /* and it isn't on the network */
	    	     !check_ip(ret, ifid)) {
			return ret;
			break;
		}
	}
	return 0;
}

/* check is an IP is taken, if it is, add it to the lease table */
int check_ip(u_int32_t addr, int ifid)
{
	struct in_addr temp;
#ifdef RONSCODE
	char hostname[MAXHOSTNAMELEN]="";
	char cid[16]="";
#endif	
	
	if (arpping(addr, server_config[ifid].server, server_config[ifid].arp, server_config[ifid].interface) == 0) {
		temp.s_addr = addr;
	 	LOG(LOG_INFO, "%s belongs to someone, reserving it for %ld seconds", 
	 		inet_ntoa(temp), server_config[ifid].conflict_time);
#ifdef RONSCODE
		add_lease(blank_chaddr, addr, server_config[ifid].conflict_time, ifid,hostname,cid);
#else
		add_lease(blank_chaddr, addr, server_config[ifid].conflict_time, ifid);
#endif
		return 1;
	} else 
	{
		return 0;
	}
}

